import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'interceptor/interceptor.dart';
import 'exceptions/response_exception.dart';
import 'request_option.dart';
import 'response.dart';

class Request {
  static Interceptor interceptor = new Interceptor();
  static bool responseFailThrow = true;
  static bool parseResponse = true;
  static Map<String, Object> defaultOptions = {
    "url": null,
    "method": RequestOption.METHOD_GET,
    "headers": {"Accept": "application/json", "Content-Type": ContentType.JSON.toString()},
    "params": {},
    "data": {},
    "meta": {},
  };

  static Future<Response> request(RequestOption option) {
    return getConfig(option).then((config) {
      HttpClient client = new HttpClient();
      String url = config.url;
      Uri parse = Uri.parse(url);
      config.params.addAll(parse.queryParameters);

      Uri target = new Uri(scheme: parse.scheme, host: parse.host, path: parse.path, queryParameters: config.params);

      return client.openUrl(config.method, target).then((client) {
        config.headers.forEach((k, v) {
          client.headers.set(k, v);
        });
        setRequestData(client, config);
        // ignore: return_of_invalid_type_from_closure
        return client.close().then((response) {
          return parseResponseBody(response).then((responseBody) {
            if (responseFailThrow && response.statusCode != 200) {
              throw new ResponseException(response: response, responseBody: responseBody);
            }
            Response resp = new Response(response: response, responseBody: responseBody);

            return interceptor.success(resp, config);
          });
        }).catchError((e) => interceptor.fail(e, config));
      });
    });
  }

  static void setRequestData(HttpClientRequest client, RequestOption config) {
    client.contentLength = -1;
    var contentType = config.headers['Content-Type'];
    if (contentType is String && contentType.contains('application/json')) {
      client.write(JSON.encode(config.data));
    } else {
      client.write(config.data);
    }
  }

  static Future<Response> get(String url, [Map<String, Object> options]) {
    RequestOption option = getOptionByMap(url, RequestOption.METHOD_GET, {}, options);

    return request(option);
  }

  static Future<Response> delete(String url, [Map<String, Object> options]) {
    RequestOption option = getOptionByMap(url, RequestOption.METHOD_DELETE, {}, options);

    return request(option);
  }

  static Future<Response> post(String url, [dynamic data, Map<String, Object> options]) {
    RequestOption option = getOptionByMap(url, RequestOption.METHOD_POST, data, options);

    return request(option);
  }

  static Future<Response> put(String url, [dynamic data, Map<String, Object> options]) {
    RequestOption option = getOptionByMap(url, RequestOption.METHOD_PUT, data, options);

    return request(option);
  }

  static Future<RequestOption> getConfig(option) {
    var config = interceptor.config(option);
    if (config is Future) {
      return config;
    }
    return new Future.value(config);
  }

  static RequestOption getOptionByMap(String url, String method, [dynamic data, Map<String, Object> options]) {
    if (options == null) {
      options = {};
    }
    options['data'] = data;

    defaultOptions.addAll(options);
    RequestOption option = new RequestOption(
        url: url, method: method, headers: defaultOptions['headers'], params: defaultOptions['params'], data: defaultOptions['data']);
    return option;
  }

  static Future<dynamic> parseResponseBody(HttpClientResponse response) {
    if (!parseResponse) {
      return new Future.value(null);
    }

    Future<String> body = response.transform(UTF8.decoder).join();

    return body.then((str) {
      String mimeType = response.headers.contentType.mimeType;
      var data = str;
      if (mimeType == 'application/json') {
        try {
          data = JSON.decode(str);
        } catch (e) {
        }
      }

      return data;
    });
  }
}
